#!/bin/bash

          STOOGES=(
                   "Curly"
                   "Larry"
                   "Moe"
                  )
 SORCERER_MIRRORS=(
                   "http://distro.ibiblio.org/pub/linux/distributions/sorcerer"
                   "ftp://ftp.phy.bnl.gov/pub/sorcerer"
                   "ftp://sorcerer.mirrors.pair.com"
                   "http://sorcerer.mirrors.pair.com"
                  )
